# pylint: disable=unsupported-assignment-operation
import contextlib
import os
from io import StringIO
from pathlib import Path
from typing import Any, Callable, Dict, NoReturn, Optional

from alembic import command as alem_command
from alembic.config import Config
from sqlalchemy import MetaData
from sqlalchemy.engine import Engine

REPO_ROOT = Path(os.path.abspath(os.path.dirname(__file__))).parent.parent.resolve()
TEST_RESOURCE_ROOT = REPO_ROOT / "src" / "test" / "resources"
TEST_VERSIONS_ROOT = REPO_ROOT / "src" / "test" / "alembic_config" / "versions"


ALEMBIC_COMMAND_MAP: Dict[str, Callable[..., NoReturn]] = {
    "upgrade": alem_command.upgrade,
    "downgrade": alem_command.downgrade,
    "revision": alem_command.revision,
    "current": alem_command.current,
    "check": alem_command.check,
}


def build_alembic_config(engine: Engine) -> Config:
    """Populate alembic configuration from metadata and config file."""
    path_to_alembic_ini = REPO_ROOT / "alembic.ini"

    alembic_cfg = Config(path_to_alembic_ini)

    alembic_cfg.set_main_option("sqlalchemy.url", engine.url.render_as_string(hide_password=False))

    alembic_cfg.set_main_option("script_location", str((Path("src") / "test" / "alembic_config")))
    return alembic_cfg


from typing import List, Union

CompareCheckConstraintsType = Union[bool, List[str], None]


def run_alembic_command(
    engine: Engine,
    command: str,
    command_kwargs: Dict[str, Any],
    target_metadata: Optional[MetaData] = None,
    compare_check_constraints: CompareCheckConstraintsType = None,
) -> str:
    command_func = ALEMBIC_COMMAND_MAP[command]

    stdout = StringIO()

    alembic_cfg = build_alembic_config(engine)
    with engine.begin() as connection:
        alembic_cfg.attributes["connection"] = connection
        if target_metadata is not None:
            alembic_cfg.attributes["target_metadata"] = target_metadata
        if compare_check_constraints is not None:
            alembic_cfg.attributes["compare_check_constraints"] = compare_check_constraints
        with contextlib.redirect_stdout(stdout):
            command_func(alembic_cfg, **command_kwargs)
    return stdout.getvalue()



from uuid import uuid4


def normalize_whitespace(text, base_whitespace: str = " ") -> str:
    """Convert all whitespace to *base_whitespace*"""
    return base_whitespace.join(text.split()).strip()


def strip_terminating_semicolon(sql: str) -> str:
    """Removes terminating semicolon on a SQL statement if it exists"""
    return sql.strip().rstrip(";").strip()


def strip_double_quotes(sql: str) -> str:
    """Removes starting and ending double quotes"""
    sql = sql.strip().rstrip('"')
    return sql.strip().lstrip('"').strip()


def escape_colon_for_sql(sql: str) -> str:
    """Escapes colons for use in sqlalchemy.text"""
    holder = str(uuid4())
    sql = sql.replace("::", holder)
    sql = sql.replace(":", r"\:")
    sql = sql.replace(holder, "::")
    return sql


def escape_colon_for_plpgsql(sql: str) -> str:
    """Escapes colons for plpgsql for use in sqlalchemy.text"""
    holder1 = str(uuid4())
    holder2 = str(uuid4())
    holder3 = str(uuid4())
    sql = sql.replace("::", holder1)
    sql = sql.replace(":=", holder2)
    sql = sql.replace(r"\:", holder3)

    sql = sql.replace(":", r"\:")

    sql = sql.replace(holder3, r"\:")
    sql = sql.replace(holder2, ":=")
    sql = sql.replace(holder1, "::")
    return sql


def coerce_to_quoted(text: str) -> str:
    """Coerces schema and entity names to double quoted one

    Examples:
        coerce_to_quoted('"public"') => '"public"'
        coerce_to_quoted('public') => '"public"'
        coerce_to_quoted('public.table') => '"public"."table"'
        coerce_to_quoted('"public".table') => '"public"."table"'
        coerce_to_quoted('public."table"') => '"public"."table"'
    """
    if "." in text:
        schema, _, name = text.partition(".")
        schema = f'"{strip_double_quotes(schema)}"'
        name = f'"{strip_double_quotes(name)}"'
        return f"{schema}.{name}"

    text = strip_double_quotes(text)
    return f'"{text}"'


def coerce_to_unquoted(text: str) -> str:
    """Coerces schema and entity names to unquoted

    Examples:
        coerce_to_unquoted('"public"') => 'public'
        coerce_to_unquoted('public') => 'public'
        coerce_to_unquoted('public.table') => 'public.table'
        coerce_to_unquoted('"public".table') => 'public.table'
    """
    return "".join(text.split('"'))



# pylint: disable=unused-argument,invalid-name,line-too-long


from typing import Generator

from sqlalchemy import text as sql_text
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.replaceable_entity import ReplaceableEntity
from alembic_utils_extended.statement import (
    coerce_to_unquoted,
    normalize_whitespace,
)


class PGExtension(ReplaceableEntity):
    """A PostgreSQL Extension compatible with `alembic revision --autogenerate`

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **signature** - *str*: A PostgreSQL extension's name
    """

    type_ = "extension"

    def __init__(self, schema: str, signature: str):
        self.schema: str = coerce_to_unquoted(normalize_whitespace(schema))
        self.signature: str = coerce_to_unquoted(normalize_whitespace(signature))
        # Include schema in definition since extensions can only exist once per
        # database and we want to detect schema changes and emit alter schema
        self.definition: str = f"{self.__class__.__name__}: {self.schema} {self.signature}"

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create extension" statement"""
        yield sql_text(f'CREATE EXTENSION "{self.signature}" WITH SCHEMA {self.literal_schema};')

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop extension" statement"""
        cascade = "CASCADE" if cascade else ""
        yield sql_text(f'DROP EXTENSION "{self.signature}" {cascade}')

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Generates SQL equivalent to "create or replace" statement"""
        raise NotImplementedError()

    @property
    def identity(self) -> str:
        """A string that consistently and globally identifies an extension"""
        # Extensions may only be installed once per db, schema is not a
        # component of identity
        return f"{self.__class__.__name__}: {self.signature}"

    def render_self_for_migration(self, omit_definition=False) -> str:
        """Render a string that is valid python code to reconstruct self in a migration"""
        var_name = self.to_variable_name()
        class_name = self.__class__.__name__

        return f"""{var_name} = {class_name}(
    schema="{self.schema}",
    signature="{self.signature}"
)\n"""

    @classmethod
    def from_database(cls, sess, schema):
        """Get a list of all extensions defined in the db"""
        sql = sql_text(
            f"""
        select
            np.nspname schema_name,
            ext.extname extension_name
        from
            pg_extension ext
            join pg_namespace np
                on ext.extnamespace = np.oid
        where
            np.nspname not in ('pg_catalog')
            and np.nspname like :schema;
        """
        )
        rows = sess.execute(sql, {"schema": schema}).fetchall()
        db_exts = [cls(x[0], x[1]) for x in rows]
        return db_exts



# pylint: disable=unused-argument,invalid-name,line-too-long


from typing import Generator

from alembic.autogenerate.api import AutogenContext
from alembic.autogenerate.render import _add_index as render_add_index
from alembic.operations.ops import CreateIndexOp
from parse import parse
from sqlalchemy import Index
from sqlalchemy import text as sql_text
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.exceptions import SQLParseFailure
from alembic_utils_extended.replaceable_entity import ReplaceableEntity
from alembic_utils_extended.statement import (
    coerce_to_unquoted,
    normalize_whitespace,
    strip_terminating_semicolon,
)


class PGMaterializedView(ReplaceableEntity):
    """A PostgreSQL Materialized View compatible with `alembic revision --autogenerate`

    Limitations:
        Materialized views may not have other views or materialized views that depend on them.

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **signature** - *str*: A SQL view's call signature
    * **definition** - *str*: The SQL select statement body of the view
    * **with_data** - *bool*: Should create and replace statements populate data
    * **indexes** - *list[Index]*: Specifies SqlAlchemy indexes to be created on MaterializedView.
    """

    type_ = "materialized_view"

    def __init__(self, schema: str, signature: str, definition: str, with_data: bool = True, indexes: list[Index] = None):
        self.schema: str = coerce_to_unquoted(normalize_whitespace(schema))
        self.signature: str = coerce_to_unquoted(normalize_whitespace(signature))
        self.definition: str = strip_terminating_semicolon(definition)
        self.with_data = with_data

        # We must specify name so that index.table.name can access this value.
        self.name = self.signature
        self.indexes = indexes

    @classmethod
    def from_sql(cls, sql: str) -> "PGMaterializedView":
        """Create an instance from a SQL string"""

        # Strip optional semicolon and all whitespace from end of definition
        # because the "with data" clause is optional and the alternative would be to enumerate
        # every possibility in the templates
        sql = strip_terminating_semicolon(sql)

        templates = [
            # Enumerate maybe semicolon endings
            "create{}materialized{}view{:s}{schema}.{signature}{:s}as{:s}{definition}{:s}with{:s}data",
            "create{}materialized{}view{:s}{schema}.{signature}{:s}as{:s}{definition}{}with{:s}{no_data}{:s}data",
            "create{}materialized{}view{:s}{schema}.{signature}{:s}as{:s}{definition}",
        ]

        for template in templates:
            result = parse(template, sql, case_sensitive=False)

            if result is not None:
                with_data = not "no_data" in result.named

                # If the signature includes column e.g. my_view (col1, col2, col3) remove them
                signature = result["signature"].split("(")[0]

                return cls(
                    schema=result["schema"],
                    # strip quote characters
                    signature=signature.replace('"', ""),
                    definition=result["definition"],
                    with_data=with_data,
                )

        raise SQLParseFailure(f'Failed to parse SQL into PGView """{sql}"""')

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create view" statement"""

        # Remove possible semicolon from definition because we're adding a "WITH DATA" clause
        definition = self.definition.rstrip().rstrip(";")

        yield sql_text(
            f'CREATE MATERIALIZED VIEW {self.literal_schema}."{self.signature}" AS {definition} WITH {"NO" if not self.with_data else ""} DATA;'
        )

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop view" statement"""
        cascade = "cascade" if cascade else ""
        yield sql_text(f'DROP MATERIALIZED VIEW {self.literal_schema}."{self.signature}" {cascade}')

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create or replace view" statement"""
        yield sql_text(f"""DROP MATERIALIZED VIEW IF EXISTS {self.literal_schema}."{self.signature}"; """)
        yield from self.to_sql_statement_create()

    def render_self_for_migration(self, omit_definition=False) -> str:
        """Render a string that is valid python code to reconstruct self in a migration"""
        var_name = self.to_variable_name()
        class_name = self.__class__.__name__
        escaped_definition = self.definition if not omit_definition else "# not required for op"

        return f"""{var_name} = {class_name}(
            schema="{self.schema}",
            signature="{self.signature}",
            definition={repr(escaped_definition)},
            with_data={repr(self.with_data)}
        )\n\n"""

    def render_post_create_entity(self, autogen_context: AutogenContext) -> str:
        index_ops = []
        for index in self.indexes if self.indexes else []:
            index.table = self
            index_op = render_add_index(autogen_context, CreateIndexOp.from_index(index))
            index_ops.append(index_op)

        return "\n" + "\n".join(index_ops) + "\n" if index_ops else ""

    @classmethod
    def from_database(cls, sess, schema):
        """Get a list of all functions defined in the db"""
        sql = sql_text(
            f"""
        select
            schemaname schema_name,
            matviewname view_name,
            definition,
            ispopulated is_populated
        from
            pg_matviews
        where
            schemaname not in ('pg_catalog', 'information_schema')
            and schemaname::text like '{schema}';
        """
        )
        rows = sess.execute(sql).fetchall()
        db_views = [cls(x[0], x[1], x[2], with_data=x[3]) for x in rows]

        for view in db_views:
            assert view is not None

        return db_views



class AlembicUtilsException(Exception):
    """Base exception for AlembicUtils package"""


class SQLParseFailure(AlembicUtilsException):
    """An entity could not be parsed"""


class FailedToGenerateComparable(AlembicUtilsException):
    """Failed to generate a comparable entity"""


class UnreachableException(AlembicUtilsException):
    """An exception no one should ever see"""


class BadInputException(AlembicUtilsException):
    """Invalid user input"""



from __future__ import annotations

import logging
from typing import Dict, List, Optional, Set, Union

from alembic.autogenerate import comparators, renderers
from alembic.autogenerate.api import AutogenContext
from alembic.operations import ops
from sqlalchemy import CheckConstraint, Enum
from sqlalchemy.engine.reflection import Inspector

logger = logging.getLogger(__name__)


def _render_create_check_constraint(
    autogen_context: AutogenContext,
    op: ops.CreateCheckConstraintOp,
) -> str:
    args = [
        repr(op.constraint_name),
        repr(op.table_name),
        repr(str(op.condition)),
    ]
    if op.schema:
        args.append(f"schema={op.schema!r}")
    kw = getattr(op, "kw", {}) or {}
    for key, value in sorted(kw.items()):
        args.append(f"{key}={value!r}")
    return f"op.create_check_constraint({', '.join(args)})"


renderers._registry[(ops.CreateCheckConstraintOp, "default")] = _render_create_check_constraint


@comparators.dispatch_for("schema")
def compare_check_constraints(
    autogen_context: AutogenContext,
    upgrade_ops: ops.UpgradeOps,
    schemas: List[Optional[str]],
) -> None:
    compare_check_constraints_opt: Union[bool, List[str], None] = autogen_context.opts.get("compare_check_constraints")

    if not compare_check_constraints_opt:
        return

    connection = autogen_context.connection
    if connection is None:
        return

    inspector: Inspector = autogen_context.inspector
    target_metadata = autogen_context.metadata

    if target_metadata is None:
        return

    if isinstance(compare_check_constraints_opt, list):
        observed_schemas: Set[str] = set(compare_check_constraints_opt)
    else:
        observed_schemas = {"public"}

    for schema in observed_schemas:
        schema_to_use = schema if schema != "public" else None

        model_constraints = _get_model_check_constraints(target_metadata, schema_to_use)
        db_constraints = _get_database_check_constraints(inspector, target_metadata, schema_to_use)

        model_constraint_names = {(c["table_name"], c["name"]) for c in model_constraints}
        db_constraint_names = {(c["table_name"], c["name"]) for c in db_constraints}

        constraints_to_create = model_constraint_names - db_constraint_names
        constraints_to_drop = db_constraint_names - model_constraint_names

        for table_name, constraint_name in constraints_to_create:
            constraint_info = next(c for c in model_constraints if c["table_name"] == table_name and c["name"] == constraint_name)
            logger.info(
                "Detected CreateCheckConstraintOp for %s.%s",
                table_name,
                constraint_name,
            )
            create_op = ops.CreateCheckConstraintOp(
                constraint_name=constraint_name,
                table_name=table_name,
                condition=constraint_info["sqltext"],
                schema=schema_to_use,
            )
            upgrade_ops.ops.append(create_op)

        for table_name, constraint_name in constraints_to_drop:
            constraint_info = next(c for c in db_constraints if c["table_name"] == table_name and c["name"] == constraint_name)
            logger.info(
                "Detected DropConstraintOp for %s.%s",
                table_name,
                constraint_name,
            )
            create_op_for_reverse = ops.CreateCheckConstraintOp(
                constraint_name=constraint_name,
                table_name=table_name,
                condition=constraint_info["sqltext"],
                schema=schema_to_use,
            )
            drop_op = ops.DropConstraintOp(
                constraint_name=constraint_name,
                table_name=table_name,
                type_="check",
                schema=schema_to_use,
                _reverse=create_op_for_reverse,
            )
            upgrade_ops.ops.append(drop_op)


def _get_model_check_constraints(metadata, schema: Optional[str]) -> List[Dict[str, str]]:
    constraints = []

    for table in metadata.tables.values():
        if table.schema != schema:
            continue

        for constraint in table.constraints:
            if isinstance(constraint, CheckConstraint):
                if constraint.name is None:
                    raise ValueError(
                        f"Unnamed check constraint on table '{table.name}'. "
                        f"All check constraints must have a name for autogenerate support. "
                        f"Constraint: {constraint.sqltext}"
                    )

                if not _constraint_columns_exist_on_table(constraint, table):
                    if not _constraint_columns_exist_in_metadata(constraint, metadata):
                        raise ValueError(
                            f"Check constraint '{constraint.name}' on table '{table.name}' "
                            f"references columns that do not exist in any table. "
                            f"Constraint: {constraint.sqltext}"
                        )
                    continue

                sqltext = str(constraint.sqltext)
                constraints.append(
                    {
                        "table_name": table.name,
                        "name": constraint.name,
                        "sqltext": sqltext,
                    }
                )

    return constraints


def _get_database_check_constraints(
    inspector: Inspector,
    metadata,
    schema: Optional[str],
) -> List[Dict[str, str]]:
    constraints = []

    for table in metadata.tables.values():
        if table.schema != schema:
            continue

        enum_constraint_names = _get_enum_constraint_names(table)

        try:
            db_constraints = inspector.get_check_constraints(table.name, schema=schema)
        except NotImplementedError:
            logger.warning("Database dialect does not support get_check_constraints. " "Check constraint autogenerate is not available.")
            return []

        for c in db_constraints:
            if c.get("name") is None:
                continue

            if c["name"] in enum_constraint_names:
                continue

            constraints.append(
                {
                    "table_name": table.name,
                    "name": c["name"],
                    "sqltext": c.get("sqltext", ""),
                }
            )

    return constraints


def _get_enum_constraint_names(table) -> Set[str]:
    constraint_names = set()
    for column in table.columns:
        if isinstance(column.type, Enum) and not getattr(column.type, "native_enum", True):
            constraint_names.add(f"{table.name}_{column.name}_check")
    return constraint_names


def _constraint_columns_exist_on_table(constraint: CheckConstraint, table) -> bool:
    table_column_names = {col.name for col in table.columns}
    constraint_columns = getattr(constraint, "columns", None)
    if constraint_columns is not None and len(constraint_columns) > 0:
        for col in constraint_columns:
            if col.name not in table_column_names:
                return False
    return True


def _constraint_columns_exist_in_metadata(constraint: CheckConstraint, metadata) -> bool:
    all_column_names: Set[str] = set()
    for table in metadata.tables.values():
        for col in table.columns:
            all_column_names.add(col.name)

    constraint_columns = getattr(constraint, "columns", None)
    if constraint_columns is not None and len(constraint_columns) > 0:
        for col in constraint_columns:
            if col.name not in all_column_names:
                return False
    return True



# pylint: disable=unused-argument,invalid-name,line-too-long
from typing import Generator, List

from parse import parse
from sqlalchemy import text as sql_text
from sqlalchemy.sql.elements import TextClause

from alembic_utils_extended.exceptions import SQLParseFailure
from alembic_utils_extended.replaceable_entity import ReplaceableEntity
from alembic_utils_extended.statement import (
    escape_colon_for_plpgsql,
    escape_colon_for_sql,
    normalize_whitespace,
    strip_terminating_semicolon,
)


class PGFunction(ReplaceableEntity):
    """A PostgreSQL Function compatible with `alembic revision --autogenerate`

    **Parameters:**

    * **schema** - *str*: A SQL schema name
    * **signature** - *str*: A SQL function's call signature
    * **definition** - *str*:  The remainig function body and identifiers
    """

    type_ = "function"

    def __init__(self, schema: str, signature: str, definition: str):
        super().__init__(schema, signature, definition)
        # Detect if function uses plpgsql and update escaping rules to not escape ":="
        is_plpgsql: bool = "language plpgsql" in normalize_whitespace(definition).lower().replace("'", "")
        escaping_callable = escape_colon_for_plpgsql if is_plpgsql else escape_colon_for_sql
        # Override definition with correct escaping rules
        self.definition: str = escaping_callable(strip_terminating_semicolon(definition))

    @classmethod
    def from_sql(cls, sql: str) -> "PGFunction":
        """Create an instance instance from a SQL string"""
        template = "create{}function{:s}{schema}.{signature}{:s}returns{:s}{definition}"
        result = parse(template, sql.strip(), case_sensitive=False)
        if result is not None:
            # remove possible quotes from signature
            raw_signature = result["signature"]
            signature = "".join(raw_signature.split('"', 2)) if raw_signature.startswith('"') else raw_signature
            return cls(
                schema=result["schema"],
                signature=signature,
                definition="returns " + result["definition"],
            )
        raise SQLParseFailure(f'Failed to parse SQL into PGFunction """{sql}"""')

    @property
    def literal_signature(self) -> str:
        """Adds quoting around the functions name when emitting SQL statements

        e.g.
        'toUpper(text) returns text' -> '"toUpper"(text) returns text'
        """
        # May already be quoted if loading from database or SQL file
        name, remainder = self.signature.split("(", 1)
        return '"' + name.strip() + '"(' + remainder

    def to_sql_statement_create(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create function" statement for PGFunction"""
        yield sql_text(f"CREATE FUNCTION {self.literal_schema}.{self.literal_signature} {self.definition}")

    def to_sql_statement_drop(self, cascade=False) -> Generator[TextClause, None, None]:
        """Generates a SQL "drop function" statement for PGFunction"""
        cascade = "cascade" if cascade else ""
        template = "{function_name}({parameters})"
        result = parse(template, self.signature, case_sensitive=False)
        try:
            function_name = result["function_name"].strip()
            parameters_str = result["parameters"].strip()
        except TypeError:
            # Did not match, NoneType is not scriptable
            result = parse("{function_name}()", self.signature, case_sensitive=False)
            function_name = result["function_name"].strip()
            parameters_str = ""

        # NOTE: Will fail if a text field has a default and that deafult contains a comma...
        parameters: List[str] = parameters_str.split(",")
        parameters = [x[: len(x.lower().split("default")[0])] for x in parameters]
        parameters = [x.strip() for x in parameters]
        drop_params = ", ".join(parameters)
        yield sql_text(f'DROP FUNCTION {self.literal_schema}."{function_name}"({drop_params}) {cascade}')

    def to_sql_statement_create_or_replace(self) -> Generator[TextClause, None, None]:
        """Generates a SQL "create or replace function" statement for PGFunction"""
        yield sql_text(f"CREATE OR REPLACE FUNCTION {self.literal_schema}.{self.literal_signature} {self.definition}")

    @classmethod
    def from_database(cls, sess, schema):
        """Get a list of all functions defined in the db"""

        # Prior to postgres 11, pg_proc had different columns
        # https://github.com/candidhealth/alembic-utils-extended/issues/12
        PG_GTE_11 = """
            and p.prokind = 'f'
        """

        PG_LT_11 = """
            and not p.proisagg
            and not p.proiswindow
        """

        # Retrieve the postgres server version e.g. 90603 for 9.6.3 or 120003 for 12.3
        pg_version_str = sess.execute(sql_text("show server_version_num")).fetchone()[0]
        pg_version = int(pg_version_str)

        sql = sql_text(
            f"""
        with extension_functions as (
            select
                objid as extension_function_oid
            from
                pg_depend
            where
                -- depends on an extension
                deptype='e'
                -- is a proc/function
                and classid = 'pg_proc'::regclass
        )

        select
            n.nspname as function_schema,
            p.proname as function_name,
            pg_get_function_arguments(p.oid) as function_arguments,
            case
                when l.lanname = 'internal' then p.prosrc
                else pg_get_functiondef(p.oid)
            end as create_statement,
            t.typname as return_type,
            l.lanname as function_language
        from
            pg_proc p
            left join pg_namespace n on p.pronamespace = n.oid
            left join pg_language l on p.prolang = l.oid
            left join pg_type t on t.oid = p.prorettype
            left join extension_functions ef on p.oid = ef.extension_function_oid
        where
            n.nspname not in ('pg_catalog', 'information_schema')
            -- Filter out functions from extensions
            and ef.extension_function_oid is null
            and n.nspname = :schema
        """
            + (PG_GTE_11 if pg_version >= 110000 else PG_LT_11)
        )

        rows = sess.execute(sql, {"schema": schema}).fetchall()
        db_functions = [cls.from_sql(x[3]) for x in rows]

        for func in db_functions:
            assert func is not None

        return db_functions
